@@ -211,7 +211,7 @@ __dir2_data_entries_count(
 			ptr += be16_to_cpu(dup->length);
 		else {
 			dep = (xfs_dir2_data_entry_t *)ptr;
-			ptr += M_DIROPS(mp)->data_entsize(dep->namelen);
+			ptr += libxfs_dir2_data_entsize(mp, dep->namelen);
 		}
 	}
 	return i;
@@ -235,7 +235,7 @@ __dir2_data_entry_offset(
 			ptr += be16_to_cpu(dup->length);
 		else {
 			dep = (xfs_dir2_data_entry_t *)ptr;
-			ptr += M_DIROPS(mp)->data_entsize(dep->namelen);
+			ptr += libxfs_dir2_data_entsize(mp, dep->namelen);
 		}
 	}
 	return ptr;
@@ -350,7 +350,7 @@ dir2_block_u_count(
 
 	btp = xfs_dir2_block_tail_p(mp->m_dir_geo, block);
 	return __dir2_data_entries_count(
-			(char *)M_DIROPS(mp)->data_unused_p(block),
+			(char *)obj + mp->m_dir_geo->data_entry_offset,
 			(char *)xfs_dir2_block_leaf_p(btp));
 }
 
@@ -369,7 +369,7 @@ dir2_block_u_offset(
 	       be32_to_cpu(block->magic) == XFS_DIR3_BLOCK_MAGIC);
 	btp = xfs_dir2_block_tail_p(mp->m_dir_geo, block);
 	ptr = __dir2_data_entry_offset(
-			(char *)M_DIROPS(mp)->data_unused_p(block),
+			(char *)obj + mp->m_dir_geo->data_entry_offset,
 			(char *)xfs_dir2_block_leaf_p(btp), idx);
 	return bitize((int)(ptr - (char *)block));
 }
@@ -486,7 +486,7 @@ dir2_data_union_tag_count(
 		end = (char *)&dep->namelen + sizeof(dep->namelen);
 		if (end > (char *)obj + mp->m_dir_geo->blksize)
 			return 0;
-		tagp = M_DIROPS(mp)->data_entry_tag_p(dep);
+		tagp = libxfs_dir2_data_entry_tag_p(mp, dep);
 	}
 	end = (char *)tagp + sizeof(*tagp);
 	return end <= (char *)obj + mp->m_dir_geo->blksize;
@@ -508,7 +508,7 @@ dir2_data_union_tag_offset(
 		return bitize((int)((char *)xfs_dir2_data_unused_tag_p(dup) -
 				    (char *)dup));
 	dep = (xfs_dir2_data_entry_t *)dup;
-	return bitize((int)((char *)M_DIROPS(mp)->data_entry_tag_p(dep) -
+	return bitize((int)((char *)libxfs_dir2_data_entry_tag_p(mp, dep) -
 			    (char *)dep));
 }
 
@@ -547,7 +547,7 @@ dir2_data_u_count(
 		return 0;
 
 	return __dir2_data_entries_count(
-				(char *)M_DIROPS(mp)->data_unused_p(data),
+				(char *)data + mp->m_dir_geo->data_entry_offset,
 				(char *)data + mp->m_dir_geo->blksize);
 }
 
@@ -564,7 +564,7 @@ dir2_data_u_offset(
 	ASSERT(be32_to_cpu(data->magic) == XFS_DIR2_DATA_MAGIC ||
 	       be32_to_cpu(data->magic) == XFS_DIR3_DATA_MAGIC);
 	ptr = __dir2_data_entry_offset(
-				(char *)M_DIROPS(mp)->data_unused_p(data),
+				(char *)data + mp->m_dir_geo->data_entry_offset,
 				(char *)data + mp->m_dir_geo->blksize, idx);
 	return bitize((int)(ptr - (char *)data));
 }
@@ -585,7 +585,7 @@ dir2_data_union_size(
 		return bitize(be16_to_cpu(dup->length));
 	else {
 		dep = (xfs_dir2_data_entry_t *)dup;
-		return bitize(M_DIROPS(mp)->data_entsize(dep->namelen));
+		return bitize(libxfs_dir2_data_entsize(mp, dep->namelen));
 	}
 }
 
